#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/select.h>

#define BUFSIZE 9

/*
 * Funkcija error_fatal() ispisuje dati string, a onda i poruku o sistemskoj
 * greski i potom prekida program.
 */
static void error_fatal (char *mesg);

/*
 * Program demonstrira situaciju kada proces (u ovom slucaju parent)
 * treba da cita podatke sa dva fajl-deskriptora (u ovom slucaju su to
 * STDIN_FILENO - standardni ulaz i filedes[0] kraj pipe-a za citanje).
 * Problem je u tome sto ne znamo kada ce podaci biti dostupni.
 *
 * U ovom kodu, je prikazano legitimno resenje ovog problema. Prvo popunimo
 * odgovarajuci skup deskriptora koji nas 'zanimaju za citanje' a potom se
 * blokiramo u select pozivu sve dok podaci ne budu dostupni na bar jednom
 * od njih.
 */
int
main (void)
{

  int filedes[2];		/* Fajl-deskriptori, krajevi pipe-a. */
  pid_t pid;			/* Proces ID child-a. */
  int n;			/* Broj procitanih karaktera. */
  fd_set read_set;		/* Skup fajl-deskriptora koji nas zanimaju za citanje. */

  /*
   * Kreira se child proces. Ako kreiranje ne uspe, stampa se poruka
   * o greski.
   */
  if (pipe (filedes) == -1)
    error_fatal ("pipe");
  /*
   * Child proces prvo zatvara kraj pipe-a koji nece koristiti.
   * Potom u bafer kopira string, zaustavlja se na dve sekunde,
   * a potom prosledjuje sadrzaj bafera parent procesu koristeci
   * pipe i na kraju prekida izvrsavanje.
   */
  if ((pid = fork ()) < 0)
    error_fatal ("fork");
  else if (pid == 0){
    char child_buf[BUFSIZE];

    close (filedes[0]);
    strcpy (child_buf, "Sistemi\n");
    sleep (5);
    if (write (filedes[1], child_buf, strlen (child_buf)) != strlen (child_buf))
      error_fatal ("write");

    exit (EXIT_SUCCESS);
  }
  else {
    char parent_buf[BUFSIZE];

    /* Parent zatvara kraj pipe-a koji nece koristiti. */
    close (filedes[1]);
    
    /* Parent u beskonacnoj petlji ... */
    while (1)
      {

	/*
	 * ... postavlja fajl-deskriptore STDIN_FILENO i filedes[0] u
	 * odgovarajuci skup fajl-deskriptora.
	 */
	FD_ZERO (&read_set);
	FD_SET (filedes[0], &read_set);
	FD_SET (STDIN_FILENO, &read_set);
	
	/*
	 * Potom se blokiramo u select-pozivu sve dok na nekom od dva
	 * deskriptora ne bude dostupan (bilo da ima podataka ili je u
	 * pitanju EOF).
	 */
	if (select (filedes[0] + 1, &read_set, NULL, NULL, NULL) == -1)
	  error_fatal ("select");

	/*
	 * Posto smo se odblokirali iz poziva select, to znaci da ili
	 * postoje podaci koje mozemo da procitamo ili je u pitanju
	 * EOF. Proveravamo da li je to slucaj na fajl deskriptoru
	 * filedes[0]. Ako jeste, sve dok ima podataka, citamo ih i
	 * ispisujemo na standardni izlaz.
	 */
	/* BITNO: U naredni if blok se ulazi i ako se naidje na EOF,
	 * tj. ulazi se u svakoj iteraciji petlje od trenutka kada se
	 * stigne do EOF. Ako ne zelimo da vise ulazimo u ovaj if
	 * blok, treba izbeci postavljanje zainteresovanosti za
	 * deskriptor filedes[0] u FD_SET kada se prvi put naidje na
	 * EOF.
	 */
	if (FD_ISSET (filedes[0], &read_set)){
	  while ((n = read (filedes[0], parent_buf, sizeof (parent_buf))) > 0)
	    if (write (STDOUT_FILENO, parent_buf, n) != n)
	      error_fatal ("write");
	  if (n < 0)
	    error_fatal ("read");
	}

	/* Radimo isto sto i za prethodni fajl-deskriptor. */
	if (FD_ISSET (STDIN_FILENO, &read_set))
	  {
	    if ((n = read (STDIN_FILENO, parent_buf, sizeof (parent_buf))) > 0)
	      {
		if (write (STDOUT_FILENO, parent_buf, n) != n)
		  error_fatal ("write");
	      }
	    else
	      {
		if (n == 0)
		  break;
		else
		  error_fatal ("read");
	      }
	  }
      }
    
  }
  /* Zavrsavamo program. */
  exit (EXIT_SUCCESS);
}
  
static void error_fatal (char *mesg){
  perror (mesg);
  exit (EXIT_FAILURE);
}
